require 'uri'
require 'forwardable'
require 'rack'

module Rack
  module Radar
    extend Forwardable

    RACK_RADAR__REQUEST_METHODS = %w[
      GET
      POST
      PUT
      PATCH
      DELETE
      OPTIONS
      HEAD
      TRACE
      LINK
      UNLINK
    ].freeze

    %w[
      header headers cookies last_request last_response
      auth authorize basic_authorize digest_auth digest_authorize token_auth token_authorize
      reset_auth! reset_basic_auth! reset_digest_auth! reset_token_auth!
    ].each do |m|
      def_delegator :__r__session, m
    end

    RACK_RADAR__DEFAULT_HOST = 'rack.radar'.freeze
    RACK_RADAR__DEFAULT_ENV = {
      'REMOTE_ADDR' => '127.0.0.1',
      'HTTP_HOST' => RACK_RADAR__DEFAULT_HOST
    }.freeze
    RACK_RADAR__XHR_KEY = 'HTTP_X_REQUESTED_WITH'.freeze
    RACK_RADAR__XHR_VAL = 'XMLHttpRequest'.freeze

    # switch session
    #
    # Rack::Radar using app based sessions, that's it, creates sessions based on app __id__.
    # you can test various apps by using `app RackApp` to switch between them.
    #
    def app app = nil
      if block_given?
        @__r__app = yield
      elsif app
        @__r__app = app
      end
      @__r__app
    end

    def map *args
      @__r__base_url = args[0] if args.size > 0
      @__r__base_url
    end

    def env
      @__r__env ||= {}
    end

    def params
      @__r__params ||= {}
    end

    def session
      env["rack.session"] ||= {}
    end

    # reset session for current app.
    # everything will be reset - cookies, headers, authorizations etc.
    def reset_app!
      __r__session :reset
    end

    alias reset_browser! reset_app!

    RACK_RADAR__REQUEST_METHODS.each do |request_method|
      define_method request_method.downcase do |*args|
        request(:http, request_method, *args)
      end
      # secure
      define_method 's_%s' % request_method.downcase do |*args|
        request(:https, request_method, *args)
      end
      # xhr
      define_method '%s_x' % request_method.downcase do |*args|
        __r__with_xhr { request(:http, request_method, *args) }
      end
      # secure xhr
      define_method 's_%s_x' % request_method.downcase do |*args|
        __r__with_xhr { request(:https, request_method, *args) }
      end
    end

    def request scheme, request_method, *args
      rest, path = args.partition {|a| a.is_a?(Hash)}
      params = self.params.merge(rest[0] || {})
      env = self.env.merge(rest[1] || {})
      __r__session.invoke_request(scheme, request_method, @__r__base_url, path, params, env)
    end

    def follow_redirect!
      last_response.redirect? || raise(RuntimeError, 'Last response is not an redirect!')
      scheme = last_request.env['HTTPS'] == 'on' ? 'https' : 'http'
      request(scheme, 'GET', [last_response['Location']], {}, {
        'HTTP_REFERER' => last_request.url,
        'rack.session' => last_request.env['rack.session'],
        'rack.session.options' => last_request.env['rack.session.options']
      })
    end

    def __r__session reset = false
      (@__r__session ||= {})[app.__id__]   = RackRadarSession.new(app) if reset
      (@__r__session ||= {})[app.__id__] ||= RackRadarSession.new(app)
    end

    def __r__with_xhr
      env_xhr = env.has_key?(RACK_RADAR__XHR_KEY) && env[RACK_RADAR__XHR_KEY]
      env[RACK_RADAR__XHR_KEY] = RACK_RADAR__XHR_VAL unless env_xhr == RACK_RADAR__XHR_VAL
      yield
      env[RACK_RADAR__XHR_KEY] = env_xhr if env_xhr
    end
  end
end

require 'rack/radar/cookies'
require 'rack/radar/session'
