module Rack
  module Radar
    class RackRadarSession
      include ::Rack::Radar

      attr_reader :headers, :cookies, :last_request, :last_response
      alias header headers

      def initialize app
        @app = assert_valid_app(app)
        @headers, @cookies = {}, RackRadarCookies.new
      end

      def basic_authorize user, pass
        @basic_auth = [user, pass]
      end
      alias authorize basic_authorize
      alias auth basic_authorize

      def digest_authorize user, pass
        @digest_auth = [user, pass]
      end
      alias digest_auth digest_authorize

      def token_authorize token, options = {}
        @token_auth = [token, options]
      end
      alias token_auth token_authorize

      def reset_basic_auth!
        @basic_auth = nil
      end

      def reset_digest_auth!
        @digest_auth = nil
      end

      def reset_token_auth!
        @token_auth = nil
      end

      def reset_auth!
        reset_basic_auth!
        reset_digest_auth!
        reset_token_auth!
      end

      def invoke_request scheme, request_method, base_url, path, params, env
        uri = uri(scheme, base_url, path)

        env = RACK_RADAR__DEFAULT_ENV.merge(env)
        env.update :method => request_method
        env.update :params => normalize_params(params)
        env.update 'HTTP_COOKIE' => cookies.to_s(uri)
        env.update basic_auth_header
        env.update token_auth_header

        process_request(@app, uri, env)

        if @digest_auth && @last_response.status == 401 && (challenge = @last_response['WWW-Authenticate'])
          env.update(digest_auth_header(challenge, uri.path, request_method))
          process_request(@app, uri, env)
        end

        cookies.persist(@last_response.header['Set-Cookie'], uri)

        @last_response.respond_to?(:finish) && @last_response.finish
        @last_response
      end

      def __r__session
        self
      end

      private

      def normalize_params params
        params.inject({}) do |params, (k, v)|
          k = k.to_s if k.is_a?(Numeric) || k.is_a?(Symbol)
          v = v.to_s if v.is_a?(Numeric) || v.is_a?(Symbol)
          params.merge k => v
        end
      end

      def assert_valid_app app
        return app if app.respond_to?(:call)
        raise(ArgumentError, 'app should be a valid Rack app')
      end

      def uri scheme, base_url, path
        path = path*'/'

        # base_url ignored if given path starting with a protocol or a slash
        path = [base_url, path].compact.map(&:to_s)*'/' unless path =~ %r[\A/|\A(\w+)?://]

        # removing leading slash(es)
        path.sub!(/\A\/+/, '')

        uri = ::URI.parse(path)
        # adding leading slash unless path starts with a protocol
        uri.path = '/' << uri.path unless path =~ %r[\A(\w+)?://]

        uri.host   ||= RACK_RADAR__DEFAULT_HOST
        uri.scheme ||= scheme.to_s

        # now that we have a full URI, eg. with scheme, port etc.
        # lets build a new one based on it
        ::URI.parse(uri.to_s)
      end

      def process_request app, path, env
        env = ::Rack::MockRequest.env_for(path.to_s, Hash[env])
        explicit_env = headers_to_env
        explicit_env['rack.input'] && env['REQUEST_METHOD'] == 'POST' && env.delete('CONTENT_TYPE')
        env.update explicit_env

        @last_request = ::Rack::Request.new(env)

        # initializing params. do not remove! needed for nested params to work
        @last_request.params

        status, headers, body = app.call(@last_request.env)

        @last_response = ::Rack::MockResponse.new(status, headers, body, env['rack.errors'].flush)
        body.respond_to?(:close) && body.close
      end

      def headers_to_env
        headers.keys.inject({}) do |headers, key|
          value = self.headers[key]
          if (key =~ /\A[[:upper:]].*\-?[[:upper:]]?.*?/) && (key !~ /\AHTTP_|\ACONTENT_TYPE\Z/)
            key = (key == 'Content-Type' ? '' : 'HTTP_') << key.upcase.gsub('-', '_')
          end
          headers.merge key => value
        end
      end

      def basic_auth_header
        return {} unless auth = @basic_auth
        {'HTTP_AUTHORIZATION' => 'Basic %s' % ["#{auth.first}:#{auth.last}"].pack("m*")}
      end

      def digest_auth_header challenge, uri, request_method
        params = ::Rack::Auth::Digest::Params.parse(challenge.split(" ", 2).last)
        params.update({
            "username" => @digest_auth.first,
                  "nc" => "00000001",
              "cnonce" => "nonsensenonce",
                 "uri" => uri,
              "method" => request_method,
        })
        params["response"] = MockDigestRequest.new(params).response(@digest_auth.last)
        {'HTTP_AUTHORIZATION' => 'Digest ' << params.map {|p| '%s="%s"' % p}.join(', ')}
      end

      def token_auth_header
        return {} unless auth = @token_auth
        chunks = auth[1].each_with_object(['token=%s' % auth[0].to_s.inspect]) do |(k,v),o|
          o << [key, value.to_s.inspect]*'='
        end
        {'HTTP_AUTHORIZATION' => 'Token %s' % chunks.join(', ')}
      end

      class MockDigestRequest # stolen from rack-test

        def initialize(params)
          @params = params
        end

        def method_missing(sym)
          if @params.has_key? k = sym.to_s
            return @params[k]
          end

          super
        end

        def method
          @params['method']
        end

        def response(password)
          Rack::Auth::Digest::MD5.new(nil).send :digest, self, password
        end

      end

    end
  end
end
